package de.kreawi.trainer;

import java.awt.BorderLayout;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.net.URI;

import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.SwingUtilities;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.kreawi.trainer.dialogs.MainDialog;
import de.kreawi.trainer.update.autoupdate.AutoUpdater;
import de.kreawi.trainer.util.OSDetector;
import de.kreawi.trainer.util.TrainerHtmlUtil;
import de.kreawi.trainer.util.TrainerTextResources;
import de.kreawi.trainer.util.swingx.DesktopApi;
import de.martinknauer.util.Util;
import de.martinknauer.util.Version;
import edu.stanford.ejalbert.BrowserLauncher;

public class LatestVersionHandler {

	private final TrainerModel model;
	private final TrainerMain main;
	private final TrainerFrame mainframe;

	private AutoUpdater autoUpdater;
	private String currentTimestamp;
	private Version dismissedVersion;
	private Version currentVersion;
	
	final private static Logger logger = LoggerFactory.getLogger(LatestVersionHandler.class); 

  public LatestVersionHandler(TrainerFrame mainframe, TrainerModel model, TrainerMain main) {
		this.mainframe = mainframe;
		this.model = model;
		this.main = main;

		init();
	}

	private void init() {
		this.autoUpdater = null;

		try {
      final boolean updateFullVerisonData = !KreawiGlobals.dataExists();
      this.autoUpdater = new AutoUpdater(updateFullVerisonData);
			this.autoUpdater.retrieveUpdateInfos();
		} catch (Exception e) {
		  logger.error("Fehler beim Einlesen der Auto-Update-URL!", e);
		}

		this.currentTimestamp = null;
		this.dismissedVersion = null;

		this.currentVersion = Util.getVersion(TrainerMain.class);
	}

	public void checkForNewVersionOnStartup(final Runnable nextTask) {

	  final boolean updateWasRequested = this.model.getSettings().getBooleanValue(TrainerSettingsConstants.KEY_UPDATE);
    final boolean updateWasDataUpdate = KreawiGlobals.isFullVersionUpdate();
    if (updateWasRequested && updateWasDataUpdate && KreawiGlobals.dataExists()) {
      // Successful full version data update -> check again for updates
      // FIXME: Funktioniert nicht
      logger.info("Successful full version data update detected. Checking for further updates!");
      this.model.getSettings().setBooleanValue(TrainerSettingsConstants.KEY_UPDATE, false);
      KreawiGlobals.setFullVersionUpdate(false);
      init();
      checkForNewVersionOnStartup(nextTask);
      return;
	  } else if (this.main.isNewVersion()) {
	    logger.info("Successful update detected. ");
      
			final String[] substitutes = { Util.getVersion(TrainerMain.class)
					.toString() };
      final String title = TrainerTextResources.getString("AUTO_UPDATE_SUCCESS_TITLE");
      final String message = TrainerTextResources.getStringWithSubstitution("AUTO_UPDATE_SUCCESS_TEXT", substitutes);
      final ImageIcon icon = new ImageIcon(TrainerMain.class.getResource(MainDialog.LOGO_PATH));
      mainframe.showMessageDialog(title, message, JOptionPane.PLAIN_MESSAGE, icon, () -> nextTask.run());
		} else if (updateWasRequested) {
		  logger.info("Failed update detected. ");
      /* update flag was set, but no new version / data update detected -> update failed */

      final String downloadUrl;
      if (this.autoUpdater.isUpdateFullVersionData()) {
        downloadUrl = TrainerTextResources.getStringWithSubstitution("LATEST_VERSION_DOWNLOAD_URL_FULL_VERSION",
            new String[] {KreawiGlobals.CURRENT_YEAR, KreawiGlobals.CURRENT_YEAR});
      } else {
        downloadUrl = TrainerTextResources.getStringWithSubstitution("LATEST_VERSION_DOWNLOAD_URL",
            KreawiGlobals.CURRENT_YEAR);
      }
      final String[] substitutes = { downloadUrl };

			SwingUtilities.invokeLater(new Runnable() {
				public void run() {

					JLabel label = new JLabel(TrainerTextResources
							.getStringWithSubstitution(
									"AUTO_UPDATE_FAILED_TEXT", substitutes));
					label.addMouseListener(new MouseAdapter() {

						public void mouseClicked(MouseEvent e) {
						  try {
						    logger.info("Öffne URL: " + downloadUrl);
                DesktopApi.browse(new URI(downloadUrl));
              } catch (Exception ex) {
                logger.error("Desktop.browse hat nicht funktioniert! (url: " + downloadUrl + ")", ex);
                try {
                  (new BrowserLauncher()).openURLinBrowser(downloadUrl);
                } catch (Exception ex2) {
                  logger.error("Konnte URL nicht öffnen! (url: " + downloadUrl + ")", ex2);
                }
              }
						}
					});
          final String failedTitle = TrainerTextResources.getString("AUTO_UPDATE_FAILED_TITLE");
          final ImageIcon failedIcon = new ImageIcon(TrainerMain.class.getResource(MainDialog.LOGO_PATH));
          JOptionPane.showMessageDialog(mainframe, label, failedTitle, JOptionPane.PLAIN_MESSAGE, failedIcon);
          nextTask.run();
				}
			});

			/*
			 * reset last received message timestamp, so update messages
			 * reappear
			 */
			this.model.getSettings().setValue(TrainerSettingsConstants.KEY_LATEST_UPDATE_MESSAGE, "0");

    } else if (!this.model.getSettings().isValueSet(TrainerSettingsConstants.KEY_CHECK_FOR_UPDATE_ON_STARTUP)) {		  
			
			SwingUtilities.invokeLater(new Runnable() {

				public void run() {
					int answer = JOptionPane
							.showConfirmDialog(
									mainframe,
									TrainerTextResources
											.getString("LATEST_VERSION_PERFORM_AUTOCHECK_QUESTION_MESSAGE"),
									TrainerTextResources
											.getString("LATEST_VERSION_PERFORM_AUTOCHECK_QUESTION_TITLE"),
									JOptionPane.YES_NO_OPTION);

					if (answer == JOptionPane.YES_OPTION) {

					  model.getSettings().setBooleanValue(TrainerSettingsConstants.KEY_CHECK_FOR_UPDATE_ON_STARTUP, true);
            
            scheduleVersionCheck(nextTask);
            
					} else {

					  model.getSettings().setBooleanValue(TrainerSettingsConstants.KEY_CHECK_FOR_UPDATE_ON_STARTUP, false);
					  
					  nextTask.run();
					}
				}

			});
			

		} else if (this.model.getSettings().getBooleanValue(TrainerSettingsConstants.KEY_CHECK_FOR_UPDATE_ON_STARTUP)) {
			scheduleVersionCheck(nextTask);
		} else {
		  // if check for update is disabled by the user
		  nextTask.run();
		}
	}
	
	/*
	 * some classes used to display the latest version dialog are obviously not
	 * thread safe, so <code>showLatestVersionDialog(..)</code> has to executed
	 * in the AWT event dispatching thread.
	 */
	private void scheduleVersionCheck(final Runnable nextTask) {
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				showLatestVersionDialog(true, nextTask);
			}
		});
	}

	public void showLatestVersionDialog() {
		showLatestVersionDialog(false, null);
	}
	
	private void showLatestVersionDialog(final boolean autoCheck, final Runnable actionIfNoUpdate) {

    if (autoCheck
        && (!hasFoundVersionInfo() || ((!hasNewVersion() || hasIgnoredNewVersion()) && !hasNewBulletinMessage()))) {
      actionIfNoUpdate.run();
      return;
    }

		String text = "";
		String title = "";
		if (this.autoUpdater.isUpdateFullVersionData()) {
	    final String updateMessage;
	    if (this.autoUpdater.getUpdateMessage() != null && this.autoUpdater.getUpdateMessage().length() > 1) {
	      updateMessage = this.autoUpdater.getUpdateMessage().replaceAll("<br />", "<br>");
	    } else {
	      updateMessage = "";
	    } 
      final String downloadUrl = TrainerTextResources.getStringWithSubstitution(
          "LATEST_VERSION_DOWNLOAD_URL_FULL_VERSION", 
          new String[] { 
              KreawiGlobals.CURRENT_YEAR,
              KreawiGlobals.CURRENT_YEAR 
          });
      final String autoUpdate = getAutoUpdateAvailableString();
      final String[] substitutes = {
          KreawiGlobals.CURRENT_YEAR, 
          downloadUrl,
          downloadUrl,
          autoUpdate,
          updateMessage
      };
      text = TrainerTextResources.getStringWithSubstitution("LATEST_VERSION_FULL_VERSION_TEXT", substitutes);
      title = TrainerTextResources.getString("LATEST_VERSION_TITLE");
		} else if (!hasFoundVersionInfo() && !autoCheck) {
			String[] substitutes = {
			    KreawiGlobals.CURRENT_YEAR, 
					this.currentVersion == null ? "N/A" : this.currentVersion
							.toString(),
					TrainerTextResources.getStringWithSubstitution("LATEST_VERSION_URL", KreawiGlobals.CURRENT_YEAR),
          TrainerTextResources.getStringWithSubstitution("LATEST_VERSION_DOWNLOAD_URL", KreawiGlobals.CURRENT_YEAR) 
          };
			text = TrainerTextResources.getStringWithSubstitution(
					"LATEST_VERSION_ERROR_TEXT", substitutes);
			title = TrainerTextResources.getString("LATEST_VERSION_TITLE");
		} else if (hasNewVersion() && (!hasIgnoredNewVersion() || !autoCheck)) {

		  final String updateMessage = this.autoUpdater.getUpdateMessage() != null
      && this.autoUpdater.getUpdateMessage().length() > 1 ? this.autoUpdater
          .getUpdateMessage().replaceAll("<br />", "<br>")
          : "";

      final String autoUpdate = getAutoUpdateAvailableString();
			final String[] substitutes = {
			    KreawiGlobals.CURRENT_YEAR,
			    (this.currentVersion == null ? "N/A" : this.currentVersion.toString()), 
			    KreawiGlobals.CURRENT_YEAR,
					this.autoUpdater.getUpdateVersion().toString(),
					TrainerTextResources.getStringWithSubstitution("LATEST_VERSION_DOWNLOAD_URL", KreawiGlobals.CURRENT_YEAR),
					autoUpdate, 
					updateMessage };

			text = TrainerTextResources.getStringWithSubstitution(
					"LATEST_VERSION_TEXT", substitutes);
			title = TrainerTextResources.getString("LATEST_VERSION_TITLE");
		} else if (hasNewBulletinMessage() && autoCheck) {
			String bulletinMessage = this.autoUpdater.getBulletinMessage()
					.length() > 1 ? this.autoUpdater.getBulletinMessage()
					.replaceAll("<br />", "<br>") : "";
			text = bulletinMessage;
			title = TrainerTextResources
					.getString("LATEST_BULLETIN_MESSAGE_TITLE");
		} else if (autoCheck) {
			return;
		} else {
      text = TrainerTextResources.getStringWithSubstitution("LATEST_VERSION_GOT_LATEST_TEXT", 
          new String[] {
            KreawiGlobals.CURRENT_YEAR, 
            (this.currentVersion == null ? "N/A" : this.currentVersion.toString()) 
          });
			title = TrainerTextResources.getString("LATEST_VERSION_TITLE");
		}

		JScrollPane scrollPane = TrainerHtmlUtil.createHTMLMessageScrollPane(text);

		BorderLayout borderLayout = new BorderLayout();
		borderLayout.setVgap(15);

		JPanel panel = new JPanel(borderLayout);
		panel.setBorder(BorderFactory.createEmptyBorder(0, 0, 10, 0));
		panel.add(scrollPane, BorderLayout.CENTER);

		JCheckBox checkBox = new JCheckBox(TrainerTextResources
				.getString("AUTO_UPDATE_MESSAGE_CHECKBOX"));

		if (autoCheck) {
			panel.add(checkBox, BorderLayout.SOUTH);
		}

		if ((!KreawiGlobals.dataExists() || (hasNewVersion() && (!hasIgnoredNewVersion() || !autoCheck)))
				&& canPerformAutoUpdate()) {
			String[] options = { TrainerTextResources.getString("AUTO_UPDATE"),
					TrainerTextResources.getString("LATEST_VERSION_BACK") };

			int answer = JOptionPane.showOptionDialog(mainframe, panel,
					title,
					JOptionPane.YES_NO_OPTION, JOptionPane.PLAIN_MESSAGE,
					new ImageIcon(TrainerMain.class
							.getResource(MainDialog.LOGO_PATH)), options,
					options[0]);

			if (answer == JOptionPane.YES_OPTION) {

				checkBox.setSelected(true);

				this.main.setAutoUpdater(autoUpdater);

				JOptionPane
						.showMessageDialog(mainframe,
								TrainerTextResources
										.getString("AUTO_UPDATE_SCHEDULED"),
								TrainerTextResources
										.getString("AUTO_UPDATE_LONG"),
								JOptionPane.PLAIN_MESSAGE,
								new ImageIcon(TrainerMain.class
										.getResource(MainDialog.LOGO_PATH)));

			}

		} else {
			JOptionPane.showMessageDialog(mainframe, panel, title,
					JOptionPane.PLAIN_MESSAGE, new ImageIcon(TrainerMain.class
							.getResource(MainDialog.LOGO_PATH)));
		}

		if (checkBox.isSelected()) {
			/* checkbox can only be selected in autocheck */
			if (hasNewVersion() && !hasIgnoredNewVersion()) {
        final String updateVersion = this.autoUpdater.getUpdateVersion().toNumericalString();
        this.model.getSettings().setValue(TrainerSettingsConstants.KEY_LATEST_UPDATE_MESSAGE, updateVersion);
			} else {
				final String bulletinTimestamp = this.autoUpdater.getBulletinTimestamp();
				this.model.getSettings().setValue(TrainerSettingsConstants.KEY_LATEST_BULLETIN_MESSAGE, bulletinTimestamp);
			}
		}

	}

  private String getAutoUpdateAvailableString() {
    final String autoUpdate;
    if (canPerformAutoUpdate()) {
      autoUpdate = TrainerTextResources.getString("AUTO_UPDATE_AVAILABLE");
    } else {
      if (isAutoUpdateAvailable()) {
        autoUpdate = TrainerTextResources.getString("AUTO_UPDATE_AVAILABLE_NO_PERMISSION");
      } else {
        // Kein Auto-Update verfügbar (z.B. Mac OS Version mit integriertem Java
        autoUpdate = "";
      }
    }
    return autoUpdate;
  }

	private boolean hasFoundVersionInfo() {
		return this.autoUpdater != null
				&& this.autoUpdater.getUpdateVersion() != null;
	}

	private boolean hasNewBulletinMessage() {

		if (this.currentTimestamp == null) {
			this.currentTimestamp = this.model.getSettings().getValue(TrainerSettingsConstants.KEY_LATEST_BULLETIN_MESSAGE);

			this.currentTimestamp = this.currentTimestamp == null ? "0"
					: this.currentTimestamp;
		}

		return this.autoUpdater.getBulletinMessage() != null
		    && !"".equals(this.autoUpdater.getBulletinMessage())
				&& this.autoUpdater.getBulletinTimestamp() != null
				&& this.autoUpdater.getBulletinTimestamp().compareTo(
						this.currentTimestamp) > 0;
	}

	private boolean hasNewVersion() {

		if (this.currentVersion == null) {
		  logger.info("Unable to determine current version for latest version check!");
			return true;
		}

		return this.autoUpdater.getUpdateVersion() != null
				&& this.autoUpdater.getUpdateVersion().compareTo(
						this.currentVersion) > 0;

	}

	private boolean hasIgnoredNewVersion() {

		if (this.dismissedVersion == null) {
			try {
        this.dismissedVersion = new Version(
            this.model.getSettings().getValue(TrainerSettingsConstants.KEY_LATEST_UPDATE_MESSAGE));
			} catch (IllegalArgumentException e) {
				/* do nothing */
			}
		}

		return this.dismissedVersion != null
				&& this.dismissedVersion.equals(this.autoUpdater
						.getUpdateVersion());

	}

	private boolean isAutoUpdateAvailable() {
		return this.autoUpdater.isAutoUpdateAvailable();
	}

	private boolean canPerformAutoUpdate() {
		return isAutoUpdateAvailable()
				&& (AutoUpdater.hasAutoUpdatePermissions() || OSDetector.getOs().isWindows());
	}

}
