package de.kreawi.trainer;

import java.awt.Color;
import java.awt.Component;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.DisplayMode;
import java.awt.desktop.QuitStrategy;
import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;

import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.UIDefaults;
import javax.swing.UIManager;
import javax.swing.plaf.ColorUIResource;

import org.rauschig.jarchivelib.ArchiveFormat;
import org.rauschig.jarchivelib.Archiver;
import org.rauschig.jarchivelib.ArchiverFactory;
import org.rauschig.jarchivelib.CompressionType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.jgoodies.looks.Options;
import com.jgoodies.looks.plastic.PlasticLookAndFeel;
import com.jgoodies.looks.plastic.PlasticXPLookAndFeel;
import com.jgoodies.looks.plastic.theme.Silver;

import de.kreawi.trainer.dialogs.KreawiInternalFrame;
import de.kreawi.trainer.update.autoupdate.AutoUpdater;
import de.kreawi.trainer.util.OSDetector;
import de.kreawi.trainer.util.TrainerHtmlUtil;
import de.kreawi.trainer.util.TrainerLoadSplash;
import de.kreawi.trainer.util.TrainerTextResources;
import de.kreawi.trainer.util.TrainerUtil;
import de.martinknauer.util.Util;
import de.martinknauer.util.Version;

public class TrainerMain {

	private TrainerModel model;

	private AutoUpdater autoUpdater;

  private TrainerFrame m_mainframe;

  private TrainerLoadSplash loadSplash;
  
  // Don't init logger before setting trainerAppDir for logging config!
  private static Logger logger; 

	public TrainerMain() {
	  
	  logger = LoggerFactory.getLogger(TrainerMain.class);
	  
		final long startTime = System.currentTimeMillis();
		
		final File appDir = KreawiGlobals.getApplicationDirectory();
    
    // Old logger necessary because of jquizengine (its using the logger!!!)
    de.martinknauer.util.logging.Logger.createInstance(false, appDir);

    logger.info("Starting trainer on " + (new SimpleDateFormat("dd.MM.yyyy HH:mm")).format(new Date(startTime)));
    
    final TrainerMain trainerMain = this;
    
    final SwingWorker<TrainerFrame, Void> worker = new SwingWorker<TrainerFrame, Void>() {

      @Override
      protected TrainerFrame doInBackground() throws Exception {
        logger.info("Trainer version "
            + (Util.getVersion(TrainerMain.class) == null ? "N/A" : Util
                .getVersion(TrainerMain.class))
            + (KreawiGlobals.isDemo() ? " running in demo mode" : ""));
        logger.info("Working directory: " + System.getProperty("user.dir"));
        logger.info("User home directory: " + System.getProperty("user.home"));
        logger.info("Application data directory: " + KreawiGlobals.getApplicationDirectory().getAbsolutePath());
        logger.info("JRE vendor: " + System.getProperty("java.vendor"));
        logger.info("JRE version: " + System.getProperty("java.version"));
        logger.info("OS: " + System.getProperty("os.name") + ", version " + System.getProperty("os.version"));
        logger.info("JVM Bit size: ");
        logger.info("  sun.arch.data.model: " + System.getProperty("sun.arch.data.model"));
        logger.info("  os.arch: " + System.getProperty("os.arch"));

        if (Util.getVersion(TrainerMain.class) == null) {
          logger.info("Starting trainer from without of a jar...");
        }
        
        model = new TrainerModel();
        
        m_mainframe = new TrainerFrame(trainerMain, model);
        
        if ((System.getProperty("os.name").startsWith("Windows") || System
            .getProperty("os.name").startsWith("Linux"))
            && model.getConfig().getBool("useCustomL&F")) {
          configureCustomUI();
        } else {
          KreawiGlobals.configureDefaultUI();
          logger.info("Custom L&F will not be used. L&F used instead: " + UIManager.getLookAndFeel().getName());
        }

        if (KreawiGlobals.getCurrentColorDepth() <= 16
            && !(KreawiGlobals.getCurrentColorDepth() == DisplayMode.BIT_DEPTH_MULTI)
            && !UIManager.getLookAndFeel().getName().contains(
                "MacLookAndFeel")) {
          UIManager.put(KreawiInternalFrame.USE_GRADIENT_BACKGROUND_KEY,
              Boolean.FALSE);
          UIManager.put(
              KreawiInternalFrame.USE_LOW_COLOR_DEPTH_BACKGROUND_KEY,
              Boolean.TRUE);
          logger.info("Using color scheme for low color depth (" + KreawiGlobals.getCurrentColorDepth() + " bit)!");
        }

        logger.info("L&F in use: " + UIManager.getLookAndFeel().getName());

        /* locale must be set before initialising the frame! */
        Locale.setDefault(new Locale(model.getConfig().getString("language")));
        logger.info("Default locale set to " + Locale.getDefault());
        
        UIManager.put("OptionPane.cancelButtonText", TrainerTextResources.getString("BUTTON_CANCEL"));
        UIManager.put("OptionPane.noButtonText", TrainerTextResources.getString("BUTTON_NO"));
        UIManager.put("OptionPane.okButtonText", TrainerTextResources.getString("BUTTON_OK"));
        UIManager.put("OptionPane.yesButtonText", TrainerTextResources.getString("BUTTON_YES"));
        
        if (Util.isExtraLargeFontSize()) {
          logger.info("Extra large font size detected.");
        } else if (Util.isLargeFontSize()) {
          logger.info("Big font size detected.");
        }
        
        logger.info("Default font size: " + Util.getSystemDefaultFontSize());

        if (KreawiGlobals.isDevMode() || KreawiGlobals.isBetaMode()) {
          final File workDir = new File(System.getProperty("user.dir"));
          if (!workDir.exists()) {
            logger.info("Kein Zugriff auf Arbeitsverzeichnis '" + workDir.getAbsolutePath() + "'!");
            File f = workDir.getParentFile();
            while (!f.exists()) {
              logger.info("Kein Zugriff auf Überverzeichnis '" + f.getAbsolutePath() + "'!");
              f = f.getParentFile();
            }
          } else {
            logger.info("Arbeitsverzeichnis existiert.");
          }
          final File dataDir = new File(workDir, KreawiGlobals.QUESTION_DIR_NAME);
          if (!dataDir.exists()) {
            logger.info("Kein Zugriff auf Fragenverzeichnis '" + dataDir.getAbsolutePath() + "'!");
          } else {
            logger.info("Fragenverzeichnis existiert.");
          }
          final File dataFile = new File(dataDir, KreawiGlobals.QUESTION_FILE_NAME);
          if (!dataFile.exists()) {
            logger.info("Kein Zugriff auf Fragendatenbank '" + dataFile.getAbsolutePath() + "'!");
          } else {
            logger.info("Fragendatenbank existiert.");
          }
          if (!appDir.exists()) {
            logger.info("Kein Zugriff auf Benutzerdatenverzeichnis '" + appDir.getAbsolutePath() + "'!");
          } else {
            logger.info("Benutzerdatenverzeichnis existiert.");
          }
          final String enabled = System.getProperty("SandboxEnabled");
          logger.info("SandboxEnabled: " + enabled);
          logger.info("LibraryDirectory: " + System.getProperty("LibraryDirectory"));
          logger.info("DocumentsDirectory: " + System.getProperty("DocumentsDirectory"));
          logger.info("CachesDirectory: " + System.getProperty("CachesDirectory"));
          logger.info("ApplicationSupportDirectory: " + System.getProperty("ApplicationSupportDirectory"));
        }

        SwingUtilities.invokeAndWait(new Runnable() {
          public void run() {
            m_mainframe.buildInterface();
          }
        });

        return m_mainframe;
      }
      
      public void done() {
        
        try {
          final TrainerFrame mainframe = get();
          mainframe.setVisible(true);

          final long startupTime = System.currentTimeMillis() - startTime;
          logger.info("Startup time: " + startupTime + "ms");
          
          if (!KreawiGlobals.isDemo()) {
            KreawiGlobals.updateAverageStartupTimeMillis(startupTime);
          }
          
        } catch (InterruptedException e) {
          logger.error("Laden des Trainers unterbrochen!", e);
        } catch (ExecutionException e) {
          final String errorMessage = TrainerTextResources.getString("APPLICATION_FATAL_ERROR_WITH_QUIT_TEXT");
          logger.error("Fehler beim Initialisieren!", e.getCause());
          final String title = TrainerTextResources.getString("APPLICATION_TITLE_SHORT");
          showMessageDialog(null, title, errorMessage, JOptionPane.ERROR_MESSAGE, null, () -> {
            finalShutdown();
          });
        } catch (CancellationException e) {
          logger.info("Laden des Trainers abgebrochen!");
        }
      }
      
    };
    
    final long averageStartupTimeMillis = KreawiGlobals.getAverageStartupTimeMillis();
    this.loadSplash = new TrainerLoadSplash(averageStartupTimeMillis, KreawiGlobals.CURRENT_YEAR, worker);
    
    SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        loadSplash.setVisible(true);
      }
    });
	}

  private static void configureCustomUI() {

		UIManager.put(Options.USE_SYSTEM_FONTS_APP_KEY, Boolean.TRUE);
		Options.setDefaultIconSize(new Dimension(18, 18));

		UIManager.put(KreawiInternalFrame.USE_GRADIENT_BACKGROUND_KEY,
				Boolean.TRUE);

		PlasticLookAndFeel.setPlasticTheme(new Silver() {
			@Override
			public ColorUIResource getFocusColor() {
				return new ColorUIResource(Color.GRAY);
			}
		});
		
		try {
			UIManager.setLookAndFeel(new PlasticXPLookAndFeel());
		} catch (Exception e) {
		  logger.error("Can't set look & feel: ", e);
		}

	}

	public void exit(final boolean forcedShutdown) {
	  
	  if (!forcedShutdown) {
	    
	    if (this.autoUpdater != null) {
	      this.model.getSettings().setBooleanValue(TrainerSettingsConstants.KEY_UPDATE, true);
	      if (this.autoUpdater.isUpdateFullVersionData()) {
	        KreawiGlobals.setFullVersionUpdate(true);
	      }
	    } else {
	      this.model.getSettings().setBooleanValue(TrainerSettingsConstants.KEY_UPDATE, false);
	      KreawiGlobals.setFullVersionUpdate(false);
	    }
	    
	    final String version;
	    if (Util.getVersion(TrainerMain.class) != null) {
	      version = Util.getVersion(TrainerMain.class).toString();
	    } else {
	      version = "1.0.0.0";
	    }
	    this.model.getSettings().setValue(TrainerSettingsConstants.KEY_VERSION, version);
	    
	    KreawiGlobals.setInstallDir(System.getProperty("user.dir"));
	  }
    

	  final Runnable continueAfterOnlineGameLogout = () -> {
	    
	    this.model.getStorage().closeDb();
	    
	    if (!forcedShutdown && this.autoUpdater != null) {
	      updateAutoupdater();
	    } else {
	      finalShutdown();
	    }
	  };
	  
	  if (!forcedShutdown && this.model.getOnlineGame().isLoggedIn()) {

	    final Runnable onSuccess = () -> {
	       logger.info("Successfully logged out from online game.");
	       continueAfterOnlineGameLogout.run();
	    };
	    final Runnable onFailure = () -> {
	      logger.warn("Error on logging out from online game!");
	      continueAfterOnlineGameLogout.run();
	    };
	    logger.info("Logging out from online game...");
	    this.model.getOnlineGame().logout(onSuccess, onFailure);

	  } else {
	    continueAfterOnlineGameLogout.run();
	  }
	}

  private void updateAutoupdater() {
    if (OSDetector.getOs().isWindows() && !AutoUpdater.hasAutoUpdatePermissions()) {
      logger.info("Can't write to program directory!");
      exitWithUpdate(null);
    } else {
      final SwingWorker<File, Void> worker = new SwingWorker<File, Void>() {
        @Override
        protected File doInBackground() throws Exception {
          try {
            File updater = autoUpdater.performAutoupdaterUpdate(this);
            if (OSDetector.getOs().isMac() && updater != null) {
              Archiver archiver = ArchiverFactory.createArchiver(ArchiveFormat.TAR, CompressionType.GZIP);
              archiver.extract(updater, updater.getParentFile());
              updater.delete();
              updater = new File(updater.getParentFile(), AutoUpdater.MAC_UPDATER_APP_NAME);
              // Execution flag should have been preserved in tar file, but it wasn't?!
              Runtime.getRuntime().exec(new String[] {"chmod", "+x", new File(updater, "Contents/MacOS/KreawiTrainerUpdater").getAbsolutePath()}, null, updater.getParentFile());
            }
            return updater;
          } catch (IOException e) {
            logger.error("Error on unpacking Mac OS Update!", e);
            return null;
          }
        }
        
        @Override
        protected void done() {
          super.done();
          File updater = null;
          try {
            updater = get();
          } catch (InterruptedException e) {
            logger.error("Updater-Download abgebrochen!", e);
          } catch (ExecutionException e) {
            logger.error("Error on Updater download!", e);
          } catch (CancellationException e) {
            logger.error("Updater-Download abgebrochen!", e);
          }
          if (updater != null) {
            logger.info("Updater was updated successfully!");
          } else {
            logger.info("Updater could not be updated!");
          }
          exitWithUpdate(updater);
        }
      };
      final String message = TrainerTextResources.getString("AUTO_UPDATE_LOADING_UPDATER");
      m_mainframe.showPleaseWaitDialogWithProgress(message, worker);
    }
  }

  private void exitWithUpdate(File updater) {
    try {
      final boolean isUpdaterDownloadFailed = updater == null;
      final boolean windowsAdminUpdater = OSDetector.getOs().isWindows() && isUpdaterDownloadFailed;
            
      if (windowsAdminUpdater) {
        logger.info("Initialising windows auto update by requesting admin permissions...");
        ProcessBuilder builder = new ProcessBuilder(new String[] {
            "cmd.exe", "/C",
            AutoUpdater.AUTO_UPDATER_ADMIN_EXE_NAME + " -loadUpdater" });
        builder.start();
      } else if (OSDetector.getOs().isMac() && !isUpdaterDownloadFailed) {
        logger.info("Initialising auto update by Updater App (" + updater.getAbsolutePath() + ")...");
        Runtime.getRuntime().exec(new String[] {"open", "-na", updater.getAbsolutePath()}, null, updater.getParentFile());
      } else {
        logger.info("Initialising auto update by updater.jar...");
        if (isUpdaterDownloadFailed) {
          updater = new File(new File(System.getProperty("user.dir")), AutoUpdater.AUTO_UPDATER_JAR_NAME);
        }
        
        TrainerUtil.startJarWithCurrentJava(updater, isUpdaterDownloadFailed ? "-loadUpdater" : "");
      }
      finalShutdown();
    } catch (IOException | URISyntaxException e) {
      logger.info("Could not execute Updater! (" + e.toString() + ")");
      SwingUtilities.invokeLater(new Runnable() {
        public void run() {
          final String title = TrainerTextResources.getString("APPLICATION_TITLE_SHORT");
          final String message = TrainerTextResources.getString("AUTO_UPDATE_ERROR_UPDATER_START");
          m_mainframe.showMessageDialog(title, message, JOptionPane.ERROR_MESSAGE, new Runnable() {
            public void run() {
              finalShutdown();
            }
          });
        }
      });
    }
  }

	private void finalShutdown() {
	  logger.info("Shutting down trainer...");

    System.exit(0);
  }

  public static void main(String[] args) {
    
    if (OSDetector.getOs().isMac()) {
      // Diese Einstellungen sind auch beim Erstellen mit dem Bundler (build.xml) gesetzt
      System.getProperties().put("apple.laf.useScreenMenuBar", "true");
      System.setProperty("apple.eawt.quitStrategy", "CLOSE_ALL_WINDOWS");
      
      // Neue Art die Strategie zu setzen (seit Java 9)
      Desktop.getDesktop().setQuitStrategy(QuitStrategy.CLOSE_ALL_WINDOWS);
    }
    
    // For Log4j configuration
    System.setProperty("trainerAppDir", KreawiGlobals.getApplicationDirectory().getAbsolutePath());
    
    String sliderIconUrl = KreawiGlobals.APP_IMAGES_DIR_NAME + "/slider.png";
    final Icon icon = new ImageIcon(KreawiGlobals.class
        .getResource(sliderIconUrl));
    final UIDefaults defaults = UIManager.getDefaults();
    defaults.put("Slider.horizontalThumbIcon", icon);

    if (args.length > 0) {
			if (args[0].equals("-dev")) {
				KreawiGlobals.setDevMode(true);
      } else if (args[0].equals("-beta")) {
        KreawiGlobals.setBetaMode(true);
			}
		}

		if (KreawiGlobals.isDevMode()) {
			new TrainerMain();
		} else {

			try {
				new TrainerMain();
			} catch (Throwable t) {
				try {
				  logger.error("Unhandled exception!", t);
					System.exit(1);
				} catch (IllegalStateException e) {
					// Logger not available
					t.printStackTrace();
				}
			}

		}
	}

  public static void showMessageDialog(final Component parent, final String title, final Object message,
      final int messageType, final ImageIcon icon, final Runnable followUpAction) {
    SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        final Object messagePane;
        if (message instanceof String) {
          final String messageString = (String) message;
          if (messageString.startsWith("<html>")) {
            messagePane = TrainerHtmlUtil.createHtmlPane(messageString);
          } else {
            messagePane = new JLabel(messageString);
          }
        } else {
          messagePane = message;
        }
        if (icon != null) {
          JOptionPane.showMessageDialog(parent, messagePane, title, messageType, icon);
        } else {
          JOptionPane.showMessageDialog(parent, messagePane, title, messageType);
        }
        if (followUpAction != null) {
          followUpAction.run();
        }
      }
    });
  }

	public void setAutoUpdater(AutoUpdater autoUpdater) {
		this.autoUpdater = autoUpdater;
	}
	
	public boolean isNewVersion() {
		return this.model.getSettings().isValueSet(TrainerSettingsConstants.KEY_VERSION)
				&& Util.getVersion(TrainerMain.class) != null
				&& Util.getVersion(TrainerMain.class).compareTo(
						new Version(this.model.getSettings().getValue(TrainerSettingsConstants.KEY_VERSION))) > 0;
	}
	
}
