package de.kreawi.trainer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.jquizengine.questions.Question;
import com.jquizengine.questions.collection.QuestionPool;

import de.kreawi.trainer.db.QuestionAttributesModel;
import de.kreawi.trainer.db.TrainerSettingsModel;
import de.kreawi.trainer.dialogs.backup.TrainerBackup;
import de.kreawi.trainer.dialogs.history.HistoryItem;
import de.kreawi.trainer.exceptions.TrainerStorageException;
import de.kreawi.trainer.onlinegame.OnlineGameMain;
import de.kreawi.trainer.onlinegame.OnlineGamePersistence;
import de.kreawi.trainer.quiz.KreawiQuiz;
import de.kreawi.trainer.util.TrainerTextResources;
import de.martinknauer.help.HelpRegistry;
import de.martinknauer.util.Util;
import de.martinknauer.util.Version;
import de.martinknauer.util.config.Config;
import de.martinknauer.util.config.ConfigFile;

public class TrainerModel {

	private QuestionPool allMcQuestions;
	private QuestionPool mcQuestionsByCategories;
	private QuestionPool mcQuestionsByExams;
	private QuestionPool allTextQuestions;
	private QuestionPool textQuestionsByCategories;
	private QuestionPool textQuestionsByExams;

	private KreawiQuiz selectedQuestions;
	
  private HelpRegistry helpRegistry;

  private final TrainerStorage trainerStorage;
  
  private OnlineGameMain onlineGame;
  
  private List<NewQuizPreparedListener> newQuizPreparedListeners = new ArrayList<NewQuizPreparedListener>();
  
  private static final Logger logger = LoggerFactory.getLogger(TrainerModel.class); 

	public TrainerModel() throws TrainerStorageException {
    
	  this.trainerStorage = new TrainerStorage();
	  
    /*replace config file with default when different version is detected */
    if (isDifferentVersion()) {
      logger.info("Detected different version since last run; replacing config file with default.");
      getConfig().createExternalConfigFile(KreawiGlobals.CONFIG_FILE_NAME, KreawiGlobals.INTERNAL_CONFIG_FILE_PATH);
    }

    setHelpRegistry(new HelpRegistry(TrainerTextResources.BUNDLE_NAME, new Config(getSettings().getSettingProperties())));

    logger.info("Total number of questions: " + getAllQuestions().getQuestionCount());
    
	}

  public TrainerStorage getStorage() {
    return this.trainerStorage;
  }
  
  public TrainerSettingsModel getSettings() {
    return this.trainerStorage.getSettingsModel();
  }
  
  public QuestionAttributesModel getQuestionAttributes() {
    return this.trainerStorage.getQuestionAttributesModel();
  }

  public OnlineGameMain getOnlineGame() {
    if (this.onlineGame == null) {
      final OnlineGamePersistence onlineGamePersistence = getStorage().getOnlineGamePersistence();
      this.onlineGame = new OnlineGameMain(onlineGamePersistence);
    }
    return this.onlineGame;
  }

  public boolean isDifferentVersion() {
    final Version currentVersion = Util.getVersion(TrainerMain.class);
    
    if (currentVersion == null) {
      /* running outside of a jar */
      return false;
    }
    
    final String storedVersion = getSettings().getValue(TrainerSettingsConstants.KEY_VERSION);
    
    return !currentVersion.toString().equals(storedVersion);
  }

  public ConfigFile getConfig() {
		return getStorage().getConfig();
	}
  
	public QuestionPool getAllQuestions() {
		return getStorage().getAllQuestions();
	}

	public QuestionPool getMcQuestionsByCategories() {
	  if (mcQuestionsByCategories == null) {
      final QuestionPool allMcQuestions = getAllMcQuestions();
      mcQuestionsByCategories = allMcQuestions.getQuestionPoolForCategories(allMcQuestions.getCategories());
      // filter out all questions which are part of exams, as they should be
      // shown in exam selection exclusively
      mcQuestionsByCategories.removeAllQuestions(getMcQuestionsByExams().getQuestions());
    }
    return mcQuestionsByCategories;
	}
	
  public QuestionPool getTextQuestionsByCategories() {
    if (textQuestionsByCategories == null) {
      final QuestionPool allTextQuestions = getAllTextQuestions();
      textQuestionsByCategories = allTextQuestions.getQuestionPoolForCategories(allTextQuestions.getCategories());
      // filter out all questions which are part of exams, as they should be
      // shown in exam selection exclusively
      textQuestionsByCategories.removeAllQuestions(getTextQuestionsByExams().getQuestions());
    }
    return textQuestionsByCategories;
  }
  
  public QuestionPool getMcQuestionsByExams() {
    if (mcQuestionsByExams == null) {
      mcQuestionsByExams = getAllMcQuestions().getQuestionPoolForExams(getAllMcQuestions().getExams());
    }
    return mcQuestionsByExams;
  }
  
  public QuestionPool getTextQuestionsByExams() {
    if (textQuestionsByExams == null) {
      textQuestionsByExams = getAllTextQuestions().getQuestionPoolForExams(getAllTextQuestions().getExams());
    }
    return textQuestionsByExams;
  }
	
	public QuestionPool getAllMcQuestions() {
	  if (allMcQuestions == null) {
	    allMcQuestions = getAllQuestions().getQuestionPoolForTypes(getMcTypes());
	  }
	  return allMcQuestions;
	}

  public QuestionPool getAllTextQuestions() {
    if (allTextQuestions == null) {
      allTextQuestions = getAllQuestions().getQuestionPoolForTypes(getTextTypes());
    }
    return allTextQuestions;
  }

  public void setSelectedQuestions(final KreawiQuiz selectedQuestions) {
		
		if (selectedQuestions != null) {
		  logger.info("" + selectedQuestions);
		}
		
		this.selectedQuestions = selectedQuestions;
		
		if (this.selectedQuestions != null) {
		  notifyNewQuizPreparedListeners(this.selectedQuestions);
		}
		
	}

	public KreawiQuiz getSelectedQuestions() {
		return selectedQuestions;
	}

	public void addQuestionToIcu(final Question question) {
	  getQuestionAttributes().setIcu(question.getId(), true);
	}
	
	public void addAllQuestionsToIcu(final QuestionPool incorrect) {
		for (Question question : incorrect.getQuestions()) {
			addQuestionToIcu(question);
		}
	}

	public void removeQuestionFromIcu(Question question) {
	   getQuestionAttributes().setIcu(question.getId(), false);
	}

  public void resetIcu() {
    getQuestionAttributes().resetIcu();
  }

	public Set<Integer> getIntensiveCareUnit() {
		return getQuestionAttributes().getIntensiveCareUnit();
	}
	
	public QuestionPool getIcuQuestions() {
    return getAllQuestions().getQuestionPoolForIds(getIntensiveCareUnit());
	}
	
	public QuestionPool getIcuQuestionsMc() {
	  return getIcuQuestions().getQuestionPoolForTypes(getMcTypes());
	}

  public QuestionPool getIcuQuestionsText() {
    return getIcuQuestions().getQuestionPoolForTypes(getTextTypes());
  }

	public boolean isInIntensiveCare(final Question question) {
	  return getQuestionAttributes().isInIcu(question.getId());
	}
	
	public void putNote(final Question question, final String note) {
		getQuestionAttributes().setNote(question.getId(), note);
	}
	
	public void eraseNote(final Question question) {
	  getQuestionAttributes().setNote(question.getId(), null);
	}
	
	public String getNote(final Question question) {
		return getQuestionAttributes().getNote(question.getId());
	}
	
  public void resetNotes() {
    getQuestionAttributes().resetNotes();
  }

	public Map<Integer, String> getNotes() {
		return getQuestionAttributes().getNotes();
	}
	
	public void setCardBoxLevel(Question question, int cardBoxLevel) {
	  getQuestionAttributes().setCardBoxLevel(question.getId(), cardBoxLevel);
	}
	
	public int getCardBoxLevel(Question question) {
		int level = getQuestionAttributes().getCardBoxLevel(question.getId());

		if (level == QuestionAttributesModel.CARD_BOX_LEVEL_UNDEFINED) {
			level = getSettings().getIntValue(TrainerSettingsConstants.KEY_DEFAULT_CARD_BOX_LEVEL);
		}

		return level;
	}
	
	public int getTrueCardBoxLevel(Question question) {
		return getQuestionAttributes().getCardBoxLevel(question.getId());
	}
	
	public void increaseCardBoxLevel(Question question) {
		int level = getCardBoxLevel(question);
		
		if (level < getConfig().getInt("maxCardBoxLevel")) {
			level++;
		}
		
		setCardBoxLevel(question, level);
	}
	
	public boolean existsFullCardBox() {
	  final int maxCardBoxLevel = getSettings().getIntValue(TrainerSettingsConstants.KEY_MAX_CARD_BOX_LEVEL);
	  for (int i=0; i<=maxCardBoxLevel; i++) {
	    final Long questionCountForCardBoxLevel = getQuestionAttributes().getQuestionCountForCardBoxLevel(i);
      if (questionCountForCardBoxLevel > KreawiGlobals.getCardBoxWarnLimit(i)) {
	      return true;
	    }
	  }
    return false;
	}

  public QuestionPool getQuestionsForCardBoxLevel(final int cardBoxLevel) {
    final boolean useUnanswered = getSettings()
        .getBooleanValue(TrainerSettingsConstants.KEY_USE_UNANSWERED_FOR_CARD_BOX);
    return getQuestionsForCardBoxLevel(cardBoxLevel, useUnanswered, true);
  }
  
  private QuestionPool getQuestionsForCardBoxLevel(final int cardBoxLevel, final boolean useUnanswered,
      final boolean includeOutOfBoundsLevelsInArchive) {
    final int maxCardBoxLevel = getSettings().getIntValue(TrainerSettingsConstants.KEY_MAX_CARD_BOX_LEVEL);
    
    final int defaultCardBoxLevel = getSettings().getIntValue(TrainerSettingsConstants.KEY_DEFAULT_CARD_BOX_LEVEL);    
    
    final QuestionPool cardBox = getQuestionsForCardBoxLevelRaw(cardBoxLevel);
    /*
     * All unanswered questions do not have a level yet, so they have to
     * be treated as if they were on the default level, unless
     * configured otherwise in the settings
     */
    if (useUnanswered && cardBoxLevel == defaultCardBoxLevel) {
      QuestionPool qp = getUnansweredQuestions();
      qp.intersect(getAllQuestions());
      cardBox.addAllQuestions(qp.getQuestions());
    }
    
    /*
     * As the maximum card box level can be changed by the user, all
     * questions which have a higher level than the current max have to
     * be treated as if they were on the current maximum level
     */
    if (includeOutOfBoundsLevelsInArchive && cardBoxLevel == maxCardBoxLevel) {
      for (int i = maxCardBoxLevel + 1; i <= getConfig().getInt("maxCardBoxLevel"); i++) {
        cardBox.addAllQuestions(getQuestionsForCardBoxLevelRaw(i).getQuestions());
      }
    }
    
    /*
     * CardBox is only applied for written questions 
     */
    cardBox.intersect(getAllMcQuestions());
    
    return cardBox;
  }
	
  /** For internal use to get the questions without adding unanswered questions or
   * the questions which have a level that is bigger than the maximum level 
   * 
   * (can happen if maximum level was changed after marking question to a 
   * higher level than the new one)
   * 
   * @param cardBoxLevel
   * @return
   */
  private QuestionPool getQuestionsForCardBoxLevelRaw(int cardBoxLevel) {
    return getAllQuestions()
        .getQuestionPoolForIds(getQuestionAttributes().getQuestionIdsForCardBoxLevel(cardBoxLevel));
  }
  
	public long getCardBoxTimestamp(Question question) {
		return getQuestionAttributes().getCardBoxTimestamp(question.getId());
	}

	public QuestionPool getUnansweredQuestions() {
		/* retrieve all question which don't have a defined question attribute */
    final Collection<Integer> allOther = getQuestionAttributes().getQuestionIdsWithDefinedCardBoxLevel();
    return getAllQuestions().getQuestionPoolForIds(allOther, true);
	}
	
  public void resetCardBox() {
    getQuestionAttributes().resetCardBox();
  }
  
  public void backupUserData(final String backupTitle) throws TrainerStorageException {
    getStorage().backupUserData(backupTitle);
  }

  public TrainerBackup[] readBackups() throws TrainerStorageException {
    return getStorage().readBackups();
  }
  
  public boolean existsOlderVersionImport() {
    return getStorage().existsOlderVersionImport();
  }
  
  public void restoreBackup(final TrainerBackup backup) throws TrainerStorageException {
    getStorage().restoreBackup(backup);
  }
      
  public List<HistoryItem> getHistoryItemsMc() {
    return getStorage().getHistoryItemsMc();
  }

  public List<HistoryItem> getHistoryItemsText() {
    return getStorage().getHistoryItemsText();
  }
  
  public void writeQuizToDisk(final KreawiQuiz quiz) throws TrainerStorageException {
    getStorage().writeQuizToDisk(quiz);
  }
  
  private static Set<Question.Type> getMcTypes() {
    Set<Question.Type> types = new HashSet<Question.Type>();
    types.add(Question.Type.COMBI);
    types.add(Question.Type.MULTIPLE);
    types.add(Question.Type.SINGLE);
    return types;
  }

  private static Set<Question.Type> getTextTypes() {
    Set<Question.Type> types = new HashSet<Question.Type>();
    types.add(Question.Type.TEXT);
    return types;
  }
	
  private void setHelpRegistry(HelpRegistry helpRegistry) {
    this.helpRegistry = helpRegistry;
  }

  public HelpRegistry getHelpRegistry() {
    return helpRegistry;
  }

  public QuestionPool searchQuestions(final String searchString) throws TrainerStorageException {
    return getStorage().searchQuestions(searchString);
  }

  public void addNewQuizPreparedListener(final NewQuizPreparedListener listener) {
    this.newQuizPreparedListeners .add(listener);
  }
  
  public void notifyNewQuizPreparedListeners(final KreawiQuiz quiz) {
    for (NewQuizPreparedListener listener : newQuizPreparedListeners) {
      listener.newQuizPrepared(quiz);
    }
  }
  
  public interface NewQuizPreparedListener {
    public void newQuizPrepared(final KreawiQuiz quiz);
  }

}
