package de.kreawi.trainer.onlinegame;

import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.kreawi.trainer.onlinegame.OnlineGameModel.OnlineGameStatus;
import de.kreawi.trainer.onlinegame.entities.OnlineGameNotification;
import de.kreawi.trainer.onlinegame.entities.OnlineGameNotification.NotificationType;
import de.kreawi.trainer.onlinegame.entities.OnlineGameQuestionResult;
import de.kreawi.trainer.onlinegame.entities.OnlineGameQuiz;
import de.kreawi.trainer.onlinegame.entities.OnlineGameResults;
import de.kreawi.trainer.onlinegame.entities.OnlineGameRoundScore;
import de.kreawi.trainer.onlinegame.entities.OnlineGameTopic;
import de.kreawi.trainer.onlinegame.entities.OnlineGameUser;
import de.kreawi.trainer.onlinegame.net.FriendListReceiver;
import de.kreawi.trainer.onlinegame.net.NotificationsReceiver;
import de.kreawi.trainer.onlinegame.net.OnlineGameCommunicator.ServerFailureCallback;
import de.kreawi.trainer.onlinegame.net.OnlineGameNotificationCommunicator;
import de.kreawi.trainer.onlinegame.net.OnlineGameProfileCommunicator;
import de.kreawi.trainer.onlinegame.net.OnlineGameProfileCommunicator.ChangeStatusData;
import de.kreawi.trainer.onlinegame.net.OnlineGameProfileCommunicator.ProfileUserDataReceiver;
import de.kreawi.trainer.onlinegame.net.OnlineGameQuizCommunicator;
import de.kreawi.trainer.onlinegame.net.OnlineGameQuizCommunicator.ResultsReceiver;
import de.kreawi.trainer.onlinegame.net.OnlineGameQuizCommunicator.TopicWithQuestionsReceiver;
import de.kreawi.trainer.onlinegame.net.OnlineGameQuizCommunicator.TopicsReceiver;
import de.kreawi.trainer.onlinegame.net.OnlineGameUserListCommunicator;

public class OnlineGameMain {
  
  private static final long NOTIFICATION_CHECK_PERIOD = 7000l;

  public static final String APP_RECORD_ID = "APP_RECORD_ID";

  public static final String APP_NAME = "APP_NAME";

  public static final String DEVICE_TYPE = "PC";

  private final OnlineGameModel model;
  
  private final OnlineGameProfileCommunicator profileCommunicator;

  private final OnlineGameQuizCommunicator quizCommunicator;
  
  private final OnlineGameUserListCommunicator userListCommunicator;

  private final OnlineGameNotificationCommunicator notificationCommunicator;


  private Timer notificationsTimer = new Timer();
  
  private List<OnlineStatusChangedListener> statusChangedListeners = new ArrayList<OnlineStatusChangedListener>();
  
  private ScheduledExecutorService networkExecutor = Executors.newSingleThreadScheduledExecutor();

  private boolean loggedIn = false;

  private LoginAfterFailureTask loginAfterFailureTask = null;

//  private static OnlineGameMain s_instance;
  
  private static final Logger logger = LoggerFactory.getLogger(OnlineGameMain.class);

  private static final int RETRY_DELAY = 5;   // deley in seconds 

  public OnlineGameMain(final OnlineGamePersistence persistence) {
    
    this.model = new OnlineGameModel(persistence);
    
    this.profileCommunicator = new OnlineGameProfileCommunicator(this.model);
    this.quizCommunicator = new OnlineGameQuizCommunicator(this.model);
    this.userListCommunicator = new OnlineGameUserListCommunicator(this.model);
    this.notificationCommunicator = new OnlineGameNotificationCommunicator(this.model);
    
  }
  
  public void signup(final String email, final String userName, final String password, final Path profileImage,
      final String mobile, final Runnable onSuccess, final ServerFailureCallback onFailure) {
    
    // deviceType is required!
    // profileImage is null here and will be set in OnlineGameProfileCommunicator.signup
    final OnlineGameUser profileUser = new OnlineGameUser(null, email, userName, null, mobile, false, DEVICE_TYPE, null,
        false, false);
    
    final ProfileUserDataReceiver saveServerProfileData = saveServerProfileData(password, onSuccess);
    
    final Runnable r = () -> {
      this.profileCommunicator.signup(profileUser, password, profileImage, saveServerProfileData, onFailure);
    };
    
    this.networkExecutor.execute(r);
  }
  
  public void resetAllData() {
    this.model.resetAllData();
    
  }
  
  public void deleteProfile(final Runnable onSuccess, final ServerFailureCallback onFailure) {
    
    final Runnable onSuccessStatus = () -> {
      final Runnable r = () -> {
        
        final String userId = model.getSettings().getProfileUser().getId();
        
        final Runnable resetDataOnSuccess = () -> {
          resetAllData();
          onSuccess.run();
        };
        
        this.profileCommunicator.deactivateUser(userId, resetDataOnSuccess, onFailure);
        
      };
      
      this.networkExecutor.execute(r);
    };
    
    final Runnable onFailureStatus = () -> {
      onFailure.run("Konnte Status nicht auf \"OFFLINE\" setzen. ");
    };
    
    setStatus(OnlineGameStatus.OFFLINE, onSuccessStatus, onFailureStatus);
    
  }

  public void editProfile(final String email, final String userName, final String password,
      final Path profileImage, final String mobile, final Runnable onSuccess, final ServerFailureCallback onFailure) {
    
    final OnlineGameUser profileUser = model.getSettings().getProfileUser();
    profileUser.setEmail(email);
    profileUser.setUserName(userName);
    profileUser.setMobile(mobile);
    
    final ProfileUserDataReceiver saveServerProfileData = saveServerProfileData(password, onSuccess);
  
    final Runnable r = () -> {
      this.profileCommunicator.editProfile(profileUser, password, profileImage, saveServerProfileData, onFailure);
    };

    this.networkExecutor.execute(r);
  }

  private ProfileUserDataReceiver saveServerProfileData(final String password, final Runnable onSuccess) {
    return new ProfileUserDataReceiver() {
      
      @Override
      public void run(final OnlineGameUser userData) {
        model.getSettings().setLogin(userData.getEmail());
        model.getSettings().setPassword(password);
        model.getSettings().updateProfileUser(userData);
        onSuccess.run();
      }
    };
  }
  
  public void login(final Runnable onSuccess, final ServerFailureCallback onFailure) {
    
    final String email = model.getSettings().getLogin();
    final String password = model.getSettings().getPassword();
    
    if (getLoginAfterFailureTask() != null) {
      try {
        getLoginAfterFailureTask().cancel();
      } catch (Exception e) {
        logger.warn("Could not cancel LoginAfterFailureTask!", e);
      }
      setLoginAfterFailureTask(null);
    }
    
    final ProfileUserDataReceiver saveSessionData = new ProfileUserDataReceiver() {
      
      @Override
      public void run(final OnlineGameUser userData) {
        model.getSettings().updateProfileUser(userData);

        final OnlineGameStatus oldStatus = model.getStatus();
        
        if (userData.isOpen()) {
          model.setStatus(OnlineGameStatus.ONLINE);
        } else {
          model.setStatus(OnlineGameStatus.BUSY);
        }
        OnlineGameMain.this.loggedIn = true;
        startNotificationCheck();
        
        notifyStatusChangedListeners(oldStatus, model.getStatus());
        onSuccess.run();
      }
    };
    
    final Runnable r = () -> {
      this.profileCommunicator.login(email, password, saveSessionData, onFailure);
    };

    OnlineGameMain.this.networkExecutor.execute(r);

  }
  
  private void startNotificationCheck() {
    logger.info("Scheduling repeating notification check. ");
    notificationsTimer.schedule(createNotificationCheckTask(true), 0l);
  }

  public void triggerNotificationCheck() {
    logger.info("Scheduling single notification check.");
    this.notificationsTimer.schedule(createNotificationCheckTask(false), 500l);
  }
  
  private NotificationTimerTask createNotificationCheckTask(final boolean reschedule) {
    return new NotificationTimerTask(notificationsTimer, OnlineGameMain.this, reschedule);
  }

  public void logout(final Runnable onSuccess, final Runnable onFailure) {
    
    final Runnable setStatusOffline = new Runnable() {
      
      @Override
      public void run() {
        model.setStatus(OnlineGameStatus.OFFLINE);
        OnlineGameMain.this.loggedIn = false;
        onSuccess.run();
      }
    };
    
    final OnlineGameUser profileUser = model.getSettings().getProfileUser();
    
    if (profileUser == null) {
      // no user registered
      onSuccess.run();
      return;
    }
    
    final Runnable r = () -> {
      this.profileCommunicator.logout(profileUser, setStatusOffline, onFailure);
    };
    
    OnlineGameMain.this.networkExecutor.execute(r);
  }
  
  public boolean isLoggedIn() {
    return this.loggedIn;
  }

  public OnlineGameStatus getStatus() {
    return model.getStatus();
  }

  public void setStatus(final OnlineGameStatus newStatus, final Runnable onSuccess, final Runnable onFailure) {
    
    final boolean open = (newStatus == OnlineGameStatus.ONLINE);
    final OnlineGameUser profileUser = model.getSettings().getProfileUser();
    if (profileUser == null) {
      logger.error("setStatus called, but not signed in?! (profile user is null)");
      return;
    }
    final ChangeStatusData changeStatusData = new ChangeStatusData(profileUser.getId(), open);
    
    final Runnable changeStatus = () -> {
      final OnlineGameStatus oldStatus = model.getStatus();
      if (oldStatus != newStatus) {
        model.setStatus(newStatus);
        notifyStatusChangedListeners(oldStatus, newStatus);
        if (OnlineGameStatus.OFFLINE.equals(oldStatus) && !OnlineGameStatus.OFFLINE.equals(newStatus)) {
          startNotificationCheck();
        }
      }
      onSuccess.run();
    };

    final Runnable r = () -> {
      this.profileCommunicator.changeStatus(changeStatusData, changeStatus, onFailure);
    };

    OnlineGameMain.this.networkExecutor.execute(r);
  }

  public OnlineGameModel getModel() {
    return this.model;
  }
  
//  public void getOpenUsers(final UserListReceiver onSuccess, final Runnable onFailure) {
//    
//    this.userListCommunicator.getOpenUsers(onSuccess, onFailure);
//    
//  }
  
  public void getFriendList(final FriendListReceiver onSuccess, final Runnable onFailure) {
    
    logger.info("Retrieving friend list...");
    
    final String userId = getModel().getSettings().getProfileUser().getId();
    
    final RetryCommand command = new RetryCommand() {
      @Override
      public void execute(final Runnable retry) {
      OnlineGameMain.this.userListCommunicator.getFriendList(userId, onSuccess, retry);
      }
    };

    final Runnable r = () -> {
      retry(1, command, onFailure);
    };
    
    OnlineGameMain.this.networkExecutor.execute(r);
  }

  public void inviteUserForQuiz(final OnlineGameUser opponent, final Runnable onSuccess, final Runnable onFailure) {
    
    logger.info("Inviting user for quiz...");
    
    final String userId = getModel().getSettings().getProfileUser().getId();
    
    final RetryCommand command = new RetryCommand() {
      @Override
      public void execute(final Runnable retry) {
        quizCommunicator.inviteUserForQuiz(userId, opponent.getId(), onSuccess, retry);
      }
    };
    
    final Runnable r = () -> {
      retry(1, command, onFailure);
    };
    
    OnlineGameMain.this.networkExecutor.execute(r);
  }
  
  public void removeFriend(OnlineGameUser friend, Runnable onSuccess, Runnable onFailure) {
    
    logger.info("Removing friend...");
    
    final OnlineGameUser profileUser = getModel().getSettings().getProfileUser();
    
    final RetryCommand command = new RetryCommand() {
      @Override
      public void execute(final Runnable retry) {
        profileCommunicator.removeFriend(profileUser, friend, onSuccess, retry);
      }
    };
    
    final Runnable r = () -> {
      retry(1, command, onFailure);
    };
    
    OnlineGameMain.this.networkExecutor.execute(r);
  }

  public void acceptInvitation(final OnlineGameQuiz quiz, final Runnable onSuccess, final Runnable onFailure) {

    logger.info("Accepting invitation...");
    
    final String userId = model.getSettings().getProfileUser().getId();
    
    final String notificationId = quiz.getLastNotificationId();
    
    final RetryCommand command = new RetryCommand() {
      @Override
      public void execute(final Runnable retry) {
        quizCommunicator.acceptQuizInvitation(userId , notificationId, onSuccess, retry);
      }
    };

    final Runnable r = () -> {
      retry(1, command, onFailure);
    };
    
    OnlineGameMain.this.networkExecutor.execute(r);
  }
  
  public void rejectInvitation(final OnlineGameQuiz quiz, final Runnable onSuccess, final Runnable onFailure) {
    
    logger.info("Rejecting invitation...");
    
    final String userId = model.getSettings().getProfileUser().getId();
    
    final String notificationId = quiz.getLastNotificationId();
    
    final Runnable onSuccessWithWorkaround = () -> {
      model.removeQuiz(quiz);
      onSuccess.run();
    };
    
    final RetryCommand command = new RetryCommand() {
      @Override
      public void execute(final Runnable retry) {
        quizCommunicator.rejectQuizInvitation(userId , notificationId, onSuccessWithWorkaround, retry);
      }
    };
    
    final Runnable r = () -> {
      retry(1, command, onFailure);
    };
    
    OnlineGameMain.this.networkExecutor.execute(r);
  }
  
  public void startQuiz(final OnlineGameQuiz quiz, final Runnable onSuccess, final Runnable onFailure) {
    
    logger.info("Starting quiz...");
    
    final String userId = model.getSettings().getProfileUser().getId();
    
    final RetryCommand command = new RetryCommand() {
      @Override
      public void execute(final Runnable retry) {
        quizCommunicator.startQuiz(userId, quiz.getId(), onSuccess, retry);
      }
    };
    
    final Runnable r = () -> {
      retry(1, command, onFailure);
    };

    OnlineGameMain.this.networkExecutor.execute(r);
  }
  
  public void loadTopics(final OnlineGameQuiz quiz, final TopicsReceiver onSuccess, final Runnable onFailure) {

    logger.info("Retrieving topics...");
    
    final String userId = model.getSettings().getProfileUser().getId();
    
    final RetryCommand command = new RetryCommand() {
      @Override
      public void execute(final Runnable retry) {
        quizCommunicator.getTopics(quiz.getId(), userId, onSuccess, retry);
      }
    };
    
    final Runnable r = () -> {
      retry(1, command, onFailure);
    };
    
    OnlineGameMain.this.networkExecutor.execute(r);
  }
  
  public void submitTopicLoadQuestions(final OnlineGameQuiz quiz, final OnlineGameTopic selectedTopic, final TopicWithQuestionsReceiver onSuccess,
      final Runnable onFailure) {
   
    logger.info("Submiting topic and retrieving questions...");
    
    final String userId = model.getSettings().getProfileUser().getId();

    quizCommunicator.submitTopicGetQuestions(userId, quiz.getId(), selectedTopic.getTopicId(), onSuccess,
        onFailure);
    
  }
  
  public void loadQuestions(final OnlineGameQuiz quiz, final TopicWithQuestionsReceiver onSuccess,
      final Runnable onFailure) {
   
    logger.info("Retrieving questions...");
    
    final String userId = model.getSettings().getProfileUser().getId();

    final RetryCommand command = new RetryCommand() {
      @Override
      public void execute(final Runnable retry) {
        quizCommunicator.getQuestions(userId, quiz.getId(), onSuccess, retry);
      }
    };
    
    final Runnable r = () -> {
      retry(1, command, onFailure);
    };
    
    OnlineGameMain.this.networkExecutor.execute(r);
  }
  
  public void submitRound(final String quizId, final int roundNo, final String timeTaken, final String rightQuestions,
      final String wrongQuestions, final OnlineGameQuestionResult[] questionResults, final Runnable onSuccess,
      final Runnable onFailure) {
    
    logger.info("Submiting round...");
    
    final String userId = model.getSettings().getProfileUser().getId();
    
    final OnlineGameRoundScore roundScore = new OnlineGameRoundScore(userId, quizId, roundNo, timeTaken,
        rightQuestions, wrongQuestions, questionResults);


    final RetryCommand command = new RetryCommand() {
      @Override
      public void execute(Runnable retry) {
        quizCommunicator.submitRoundScore(roundScore, onSuccess, onFailure);
      }
    };

    final Runnable r = () -> {
      retry(1, command, onFailure);
    };
  
    OnlineGameMain.this.networkExecutor.execute(r);
  }


  public void loadQuizResults(final OnlineGameQuiz quiz, final Runnable onSuccess, final Runnable onFailure) {
    
    logger.info("Retrieving quiz results...");
    
    final String userId = model.getSettings().getProfileUser().getId();
    
    final ResultsReceiver receiveResults = new ResultsReceiver() {
      @Override
      public void run(OnlineGameResults results) {
        
        quiz.setResults(results);
        
        onSuccess.run();
      }
    };
    
    final RetryCommand command = new RetryCommand() {
      @Override
      public void execute(Runnable retry) {
        quizCommunicator.getQuizResults(userId, quiz.getId(), receiveResults , retry);
      }
    };
    
    final Runnable r = () -> {
      retry(1, command, onFailure);
    };
    
    OnlineGameMain.this.networkExecutor.execute(r);
  }
  
  private void retry(final int tryNo, final RetryCommand command, final Runnable onLastFailure) {
    
    if (tryNo > 3) {
      onLastFailure.run();
      return;
    }
    
    final Runnable retry = () -> {
      retry(tryNo + 1, command, onLastFailure);
    };
    
    if (tryNo > 1) {
      try {
        TimeUnit.SECONDS.sleep(RETRY_DELAY);
      } catch (InterruptedException e) {
        return;
      }
    }
    command.execute(retry);
  }
  
  private void checkForNotifications(final NotificationsReceiver onSuccess, final Runnable onFailure) {
    
    logger.debug("Checking for new notifications!");

    final String userId = model.getSettings().getProfileUser().getId();
    final Runnable r = () -> {
      notificationCommunicator.getNotifications(userId, onSuccess, onFailure);
    };
    
    this.networkExecutor.execute(r);
    
  }

  private class NotificationTimerTask extends TimerTask {
    
    private final Timer timer;
    private final OnlineGameMain onlineGameMain;
    private final boolean reschedule;
    
    public NotificationTimerTask(final Timer timer, final OnlineGameMain onlineGameMain, final boolean reschedule) {
      this.timer = timer;
      this.onlineGameMain = onlineGameMain;
      this.reschedule = reschedule;
    }

    @Override
    public void run() {

      if (this.onlineGameMain.getModel().getStatus() == OnlineGameStatus.OFFLINE) {
        // no execution and scheduling after logout
        logger.debug("Notification check cancelled (game is offline)!");
        return;
      }
      
      final NotificationsReceiver onSuccess = new NotificationsReceiver() {
        @Override
        public void run(List<OnlineGameNotification> current, List<OnlineGameNotification> recent) {
          if (current.size() > 0 || recent.size() > 0) {
            String cur = "";
            for (OnlineGameNotification n : current) {
              cur += " " + n.getType() + " (" + n.getMessage() + ")";
            }
            String rec = "";
            for (OnlineGameNotification n : recent) {
              rec += " " + n.getType() + " (" + n.getMessage() + ")";
            }
            logger.debug("Current notifications:" + cur + "; recent:" + rec);
            onlineGameMain.processNotifications(current, recent);
          }
          if (reschedule) {
            timer.schedule(new NotificationTimerTask(timer, onlineGameMain, true), NOTIFICATION_CHECK_PERIOD);
          }
        }
      };
      final Runnable onFinalFailure = () -> {
        if (reschedule) {
          final Runnable onSuccessfulLoguot = () -> {
            logger.info("Logout after final notification check failure successful.");
          };
          final Runnable onLogoutFailure = () -> {
            logger.info("Logout after final notification check failure also failed. Forcing offline status...");
            onlineGameMain.getModel().setStatus(OnlineGameStatus.OFFLINE);
          };
          onlineGameMain.setStatus(OnlineGameStatus.OFFLINE, onSuccessfulLoguot, onLogoutFailure);
          onlineGameMain.scheduleReloginAfterConnectionFailure();
        }
      };
      
      final RetryCommand command = new RetryCommand() {
        @Override
        public void execute(Runnable retry) {
          checkForNotifications(onSuccess, retry);
        }
      };
      
      // Scheduled on networkExecuter in checkForNotifications function
      retry(1, command, onFinalFailure);
      
    }
    
  }
  
  protected void scheduleReloginAfterConnectionFailure() {
    final long delay = 300000l;
    logger.warn("Scheduling login in " + (delay / 1000l / 60l) + " minutes...");
    setLoginAfterFailureTask(new LoginAfterFailureTask(this));
    this.notificationsTimer.schedule(getLoginAfterFailureTask(), delay);
  }

  private class LoginAfterFailureTask extends TimerTask {
    
    private final OnlineGameMain onlineGameMain;
    
    public LoginAfterFailureTask(final OnlineGameMain onlineGameMain) {
      this.onlineGameMain = onlineGameMain;
    }

    @Override
    public void run() {
      onlineGameMain.setLoginAfterFailureTask(null);
      final Runnable onSuccess = () -> {
        logger.info("Successfully logged in after final failure of notification check.");
      };
      final ServerFailureCallback onFailure = new ServerFailureCallback() {
        @Override
        public void run(String serverMessage) {
          logger.error("Failure on logging in. Message: " + serverMessage + ")");
          onlineGameMain.scheduleReloginAfterConnectionFailure();
        }
      };
      onlineGameMain.login(onSuccess, onFailure);
    }
  }

  protected LoginAfterFailureTask getLoginAfterFailureTask() {
    return loginAfterFailureTask;
  }

  protected void setLoginAfterFailureTask(LoginAfterFailureTask loginAfterFailureTask) {
    this.loginAfterFailureTask = loginAfterFailureTask;
  }

  protected void processNotifications(final List<OnlineGameNotification> current,
      final List<OnlineGameNotification> recent) {
    
    List<OnlineGameNotification> currentAndRecent = new ArrayList<OnlineGameNotification>();
    currentAndRecent.addAll(current);
    currentAndRecent.addAll(recent);
    
    for (OnlineGameNotification notification : currentAndRecent) {
      
      logger.debug("Processing notification: " + notification.toString());
      
      markNotificationRead(notification);
      
      final NotificationType type = notification.getType();

      switch (type) {
      case CHALLENGE_INVITATION:
      case CHALLENGE_REJECTED:
      case CHALLENGE_ACCEPTED:
      case CHALLENGE_INVITATION_SENT:

        logger.error("Type of notification not applicable! (Type: " + type);
        
        return;
        
      case QUIZ_INVITATION:
      case QUIZ_INVITATION_SENT:
        
        // notification may be present twice
        if (model.getCurrentQuiz(notification.getQuizChallengeId()) == null) {
          model.addNewQuiz(notification);
        }
        
        break;
      case QUIZ_ACCEPTED:
      case QUIZ_REJECTED:
      case QUIZ_CHALLENGE_PENDING:
      case SELECT_TOPIC:
      case QUIZ_CHALLENGE_STARTED:
      case QUIZ_CHALLENGE_ROUND_PLAYING:
      case QUIZ_CHALLENGE_ROUND_WAITING:
      case QUIZ_CHALLENGE_WON:
      case QUIZ_CHALLENGE_LOSE:
      case QUIZ_CHALLENGE_DRAW:
      case FINISHED:

        setQuizStatus(notification);

        break;
        
      }
      
    }
  }

  private void markNotificationRead(final OnlineGameNotification notification) {
    
    final Runnable markRead = () -> {
      
      try {
        Thread.sleep(200);
      } catch (InterruptedException e) {
        logger.error("Thread for marking notification read was unexpectedly interrupted!", e);
      }
      
      final String userId = model.getSettings().getProfileUser().getId();
      
      final Runnable onSuccess = () -> {
        logger.debug("Marked as read: " + notification.toString());
      };
      final Runnable onFailure = () -> {
        logger.error("Failure marking as read: " + notification.toString());
      };
      notificationCommunicator.updateQuizNotificationStatus(userId, notification.getNotificationId(), onSuccess,
          onFailure);
    };
    
    final Thread thread = new Thread(markRead);
    
    thread.start();
    
  }

  private void setQuizStatus(final OnlineGameNotification notification) {
    
    final String quizId = notification.getQuizChallengeId();
    
    OnlineGameQuiz quiz = model.getCurrentQuiz(quizId);
    
    if (quiz == null) {
      quiz = model.getFinishedQuiz(quizId);
    }
    
    // quiz from another client???
    if (quiz == null) {
      
      logger.warn("Tried to set status for non-existant quiz. Creating a new one! "
          + "(notification: " + notification + ")");
      
      model.addNewQuiz(notification);
      
    } else {
      
      quiz.setStatus(notification.getType());
      quiz.setLastNotificationId(notification.getNotificationId());
      model.updateQuiz(quiz);
      
    }
  }

  public List<OnlineGameQuiz> getCurrentQuizzes() {
    return model.getCurrentQuizzes();
  }

  public List<OnlineGameQuiz> getFinishedQuizzes() {
    return model.getFinishedQuizzes();
  }
  
  public void addStatusChangedListener(final OnlineStatusChangedListener listener) {
    if (!this.statusChangedListeners.contains(listener)) {
      this.statusChangedListeners.add(listener);
    }
  }
  
  public void removeStatusChangedListener(final OnlineStatusChangedListener listener) {
    this.statusChangedListeners.remove(listener);
  }
  
  private void notifyStatusChangedListeners(final OnlineGameStatus oldStatus, final OnlineGameStatus newStatus) {
    
    if (this.statusChangedListeners == null) {
      logger.warn("statusChangedListeners is null!");
      return;
    }
    
    for (OnlineStatusChangedListener listener : this.statusChangedListeners) {
      listener.onStatusChanged(oldStatus, newStatus);
    }
    
  }

  public static interface OnlineStatusChangedListener {
    public void onStatusChanged(final OnlineGameStatus oldStatus, final OnlineGameStatus newStatus);
  }

}
